(* $Id: glTex.ml,v 1.13 2008/10/25 02:22:58 garrigue Exp $ *)

open Gl
open GlPix

type env_param = [
    `mode of [`modulate|`decal|`blend|`replace] 
  | `color of rgba
]
external env : env_param -> unit = "ml_glTexEnv"

let is_pow2 n =
    n land (n - 1) = 0

let npot = ref None

let check_pow2 n =
  if !npot = None then
    npot := Some (GlMisc.check_extension "GL_OES_texture_npot");
  (!npot = Some true) || is_pow2 n

type format = [
  | `alpha
  | `rgb
  | `rgba
  | `luminance
  | `luminance_alpha
  | `bgra
]

type compressed_format = [
    `palette4_rgb8
  | `palette4_rgba8
  | `palette4_r5_g6_b5
  | `palette4_rgba4
  | `palette4_rgb5_a1
  | `palette8_rgb8
  | `palette8_rgba8
  | `palette8_r5_g6_b5
  | `palette8_rgba4
  | `palette8_rg5b_a1
  | `rgb_pvrtc_4bppv1
  | `rgb_pvrtc_2bppv1
  | `rgba_pvrtc_4bppv1
  | `rgba_pvrtc_2bppv1
]

external image2d :
    level:int -> internal:[<format] -> width:int -> height:int ->
    border:int -> format:[< format] -> [< kind] Raw.t -> unit
    = "ml_glTexImage2D_bc" "ml_glTexImage2D"
let image2d ?(level=0) ?internal:i ?(border=false) img =
  let internal = match i with None -> (format img) | Some f -> f in
  let border = if border then 1 else 0 in
  if not (check_pow2 (width img - 2 * border)) then
    raise (GLerror "GlTex.image2d : bad width");
  if not (check_pow2 (height img - 2 * border)) then
    raise (GLerror "GlTex.image2d : bad height");
  image2d ~level ~internal ~border
    ~width:(width img) ~height:(height img) ~format:(format img) (to_raw img)
external sub_image2d :
    level:int -> xoffset:int -> yoffset:int -> width:int -> height:int ->
    format:[< format] -> [< kind] Raw.t -> unit
    = "ml_glTexSubImage2D_bc" "ml_glTexSubImage2D"
let sub_image2d ?(level=0) xoffset yoffset img =
  sub_image2d ~level ~xoffset ~yoffset
    ~width:(width img) ~height:(height img) ~format:(format img) (to_raw img)
external compressed_image2d :
    level:int -> internal:[< compressed_format] -> width:int -> height:int ->
    border:int -> [`ubyte] Raw.t -> unit
    = "ml_glCompressedTexImage2D_bc" "ml_glCompressedTexImage2D"
let compressed_image2d ?(level=0) ?(border=false) internal width height raw =
    let border = if border then 1 else 0 in
    if not (is_pow2 (width - 2 * border)) then
        raise (GLerror "GlTex.compressed_image2d : bad width");
    if not (is_pow2 (height - 2 * border)) then
        raise (GLerror "GlTex.compressed_image2d : bad height");
    compressed_image2d ~level ~internal ~width ~height ~border raw

type filter = [
    `nearest
  | `linear
  | `nearest_mipmap_nearest
  | `linear_mipmap_nearest
  | `nearest_mipmap_linear
  | `linear_mipmap_linear
]
type wrap = [`repeat|`clamp_to_edge]
type parameter = [
    `min_filter of filter
  | `mag_filter of [`nearest|`linear]
  | `wrap_s of wrap
  | `wrap_t of wrap
  | `generate_mipmap of bool
] 
external parameter : target:[`texture_2d] -> parameter -> unit
    = "ml_glTexParameter"

type texture_id = nativeint
let texture0 = 0n
external _gen_textures : int -> [`uint] Raw.t -> unit = "ml_glGenTextures"
let gen_textures ~len =
  let raw = Raw.create `uint ~len in
  _gen_textures len raw;
  let arr = Array.create len Nativeint.zero in
  for i = 0 to len - 1 do
    arr.(i) <- Raw.get_long raw ~pos:i
  done;
  arr
let gen_texture () =  (gen_textures 1).(0)

external bind_texture : target:[`texture_2d] -> texture_id -> unit
    = "ml_glBindTexture"
external delete_texture : texture_id -> unit = "ml_glDeleteTexture"
let delete_textures a = Array.iter (fun id -> delete_texture id) a
