(* varcheck.ml     Find variants that aren't defined as tags
 *)
open StdLabels
open Genlex

let lexer = make_lexer ["->"; "$$"]

(* Naive definitions of character types for now.
 *)
let isgrave_o =
    function
    | None -> false
    | Some c -> c = '`'

let isspace_o =
    function
    | None -> false
    | Some c -> c = ' ' || c = '\n' || c = '\t'

let isalnum_o =
    function
    | None -> false
    | Some c ->
        (c >= 'a' && c <= 'z') ||
        (c >= 'A' && c <= 'Z') ||
        (c >= '0' && c <= '9') ||
        c = '_' || c = '\''

let getident str =
    let b = Buffer.create 32
    in let () =
        begin
        while isspace_o (Stream.peek str) do
            Stream.junk str
        done;
        while isalnum_o (Stream.peek str) do
            Buffer.add_char b (Stream.next str);
        done;
        end
    in
        Buffer.contents b


let rec gettag str =
    let skippable co =
        co <> None && not (isgrave_o co)
    in
        let () =
            begin
            while skippable (Stream.peek str) do
                Stream.junk str
            done;
            if Stream.peek str = None then
                raise End_of_file;
            Stream.junk str;
            end
        in
            getident str


let checkfile tags fn =
    let str = Stream.of_channel (open_in fn)
    in
        try
            while true do
                let tag = gettag str
                in
                    if not (Hashtbl.mem tags tag) then
                        Printf.printf "%s: unknown tag %s\n" fn tag
            done
        with End_of_file -> ()

let tags_file_name s =
    let slen = String.length s
    in
        slen > 4 && String.sub s (slen - 4) 4 = ".var"


let gettags tags fn =
  let s = lexer (Stream.of_channel (open_in fn)) in
    try while true do match s with parser
        [< ' Ident tag >] -> Hashtbl.add tags tag ()
      | [< ' Kwd "->"; ' Ident _ >] -> ()
      | [< ' Kwd "$$" >] -> ()
      | [< >] -> raise End_of_file
    done with End_of_file -> ()


let main () =
  begin
  if Array.length Sys.argv < 2 || not (tags_file_name Sys.argv.(1)) then
    begin
    Printf.fprintf stderr "varcheck: need at least one tags file\n";
    exit 1;
    end;
  let tags = Hashtbl.create 57 in
  let rec get i =
    if i < Array.length Sys.argv && tags_file_name Sys.argv.(i) then
        begin
        gettags tags Sys.argv.(i);
        get (i + 1);
        end
    else
        i
  in
  let start_ix = get 1 in
  for i = start_ix to Array.length Sys.argv - 1 do
    checkfile tags Sys.argv.(i);
  done
  end

let _ = Printexc.print main ()
